@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -32,7 +32,8 @@
 
 namespace Ovito {
 
-IMPLEMENT_OVITO_CLASS(GALAMOSTImporter);
+IMPLEMENT_CREATABLE_OVITO_CLASS(GALAMOSTImporter);
+OVITO_CLASSINFO(GALAMOSTImporter, "DisplayName", "GALAMOST");
 
 /******************************************************************************
 * Checks if the given file has format that can be read by this importer.
@@ -63,7 +64,8 @@ bool GALAMOSTImporter::OOMetaClass::checkFileFormat(const FileHandle& file) cons
 ******************************************************************************/
 void GALAMOSTImporter::FrameLoader::loadFile()
 {
-    setProgressText(tr("Reading GALAMOST file %1").arg(fileHandle().toString()));
+    TaskProgress progress(this_task::ui());
+    progress.setText(tr("Reading GALAMOST file %1").arg(fileHandle().toString()));
 
     // Initialize XML reader and open input file.
     std::unique_ptr<QIODevice> device = fileHandle().createIODevice();
@@ -120,8 +122,7 @@ void GALAMOSTImporter::FrameLoader::loadFile()
 
             // Parse the child elements.
             while(xml.readNextStartElement()) {
-                if(isCanceled())
-                    return;
+                this_task::throwIfCanceled();
 
                 if(xml.name().compare(QStringLiteral("box")) == 0) {
                     // Parse box dimensions.
