@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -57,24 +57,23 @@ MarchingCubes::MarchingCubes(SurfaceMeshBuilder& outputMesh, int size_x, int siz
 /******************************************************************************
  * Main method that constructs the isosurface mesh.
  ******************************************************************************/
-bool MarchingCubes::generateIsosurface(FloatType isolevel, ProgressingTask& operation)
+void MarchingCubes::generateIsosurface(FloatType isolevel, TaskProgress& progress)
 {
     _isolevel = isolevel;
     int size_x = _infiniteDomain ? (_size_x - 1) : _size_x;
     int size_y = _infiniteDomain ? (_size_y - 1) : _size_y;
     int size_z = _infiniteDomain ? (_size_z - 1) : _size_z;
 
-    operation.setProgressMaximum(size_z * 2);
-    computeIntersectionPoints(operation);
+    progress.setMaximum(size_z * 2);
+    computeIntersectionPoints(progress);
 
-    if(operation.isCanceled())
-        return false;
+    this_task::throwIfCanceled();
 
     if(_outputMesh.spaceFillingRegion() != SurfaceMesh::InvalidIndex) {
         handleSpaceFillingRegion();
         _vertexGrower.reset();
         _faceGrower.reset();
-        return !operation.isCanceled();
+        return;
     }
 
     // Setup region calculation
@@ -83,7 +82,7 @@ bool MarchingCubes::generateIsosurface(FloatType isolevel, ProgressingTask& oper
         _maxRegionIndex = 0;
     }
 
-    for(int k = 0; k < size_z; k++, operation.incrementProgressValue()) {
+    for(int k = 0; k < size_z; k++, progress.incrementValue()) {
         for(int j = 0; j < size_y; j++) {
             for(int i = 0; i < size_x; i++) {
                 _lut_entry = 0;
@@ -97,8 +96,6 @@ bool MarchingCubes::generateIsosurface(FloatType isolevel, ProgressingTask& oper
                 processCube(i, j, k);
             }
         }
-        if(operation.isCanceled())
-            return false;
     }
 
     if(identifyRegions()) {
@@ -116,18 +113,18 @@ bool MarchingCubes::generateIsosurface(FloatType isolevel, ProgressingTask& oper
     _vertexGrower.reset();
     _faceGrower.reset();
 
-    return !operation.isCanceled();
+    this_task::throwIfCanceled();
 }
 
 /******************************************************************************
  * Compute the intersection points with the isosurface along the cube edges.
  ******************************************************************************/
-void MarchingCubes::computeIntersectionPoints(ProgressingTask& operation)
+void MarchingCubes::computeIntersectionPoints(TaskProgress& progress)
 {
     if(_pbcFlags[0] && _pbcFlags[1] && _pbcFlags[2])
         _outputMesh.setSpaceFillingRegion(0);
 
-    for(int k = 0; k < _size_z && !operation.isCanceled(); k++, operation.incrementProgressValue()) {
+    for(int k = 0; k < _size_z; k++, progress.incrementValue()) {
         for(int j = 0; j < _size_y; j++) {
             for(int i = 0; i < _size_x; i++) {
                 FloatType cube[8];
@@ -164,50 +161,52 @@ bool MarchingCubes::testFace(signed char face)
     FloatType A, B, C, D;
 
     switch(face) {
-        case -1:
-        case 1:
-            A = _cube[0];
-            B = _cube[4];
-            C = _cube[5];
-            D = _cube[1];
-            break;
-        case -2:
-        case 2:
-            A = _cube[1];
-            B = _cube[5];
-            C = _cube[6];
-            D = _cube[2];
-            break;
-        case -3:
-        case 3:
-            A = _cube[2];
-            B = _cube[6];
-            C = _cube[7];
-            D = _cube[3];
-            break;
-        case -4:
-        case 4:
-            A = _cube[3];
-            B = _cube[7];
-            C = _cube[4];
-            D = _cube[0];
-            break;
-        case -5:
-        case 5:
-            A = _cube[0];
-            B = _cube[3];
-            C = _cube[2];
-            D = _cube[1];
-            break;
-        case -6:
-        case 6:
-            A = _cube[4];
-            B = _cube[7];
-            C = _cube[6];
-            D = _cube[5];
-            break;
-        default: OVITO_ASSERT_MSG(false, "Marching cubes", "Invalid face code");
-    };
+    case -1:
+    case 1:
+        A = _cube[0];
+        B = _cube[4];
+        C = _cube[5];
+        D = _cube[1];
+        break;
+    case -2:
+    case 2:
+        A = _cube[1];
+        B = _cube[5];
+        C = _cube[6];
+        D = _cube[2];
+        break;
+    case -3:
+    case 3:
+        A = _cube[2];
+        B = _cube[6];
+        C = _cube[7];
+        D = _cube[3];
+        break;
+    case -4:
+    case 4:
+        A = _cube[3];
+        B = _cube[7];
+        C = _cube[4];
+        D = _cube[0];
+        break;
+    case -5:
+    case 5:
+        A = _cube[0];
+        B = _cube[3];
+        C = _cube[2];
+        D = _cube[1];
+        break;
+    case -6:
+    case 6:
+        A = _cube[4];
+        B = _cube[7];
+        C = _cube[6];
+        D = _cube[5];
+        break;
+    default:
+        OVITO_ASSERT_MSG(false, "Marching cubes", "Invalid face code");
+        return false;
+    }
 
     if(std::abs(A * C - B * D) < _epsilon) return face >= 0;
     return face * A * (A * C - B * D) >= 0;  // face and A invert signs
