@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -34,8 +34,9 @@
 
 namespace Ovito {
 
-IMPLEMENT_OVITO_CLASS(ParaViewVTMFileFilter);
-IMPLEMENT_OVITO_CLASS(ParaViewVTMImporter);
+IMPLEMENT_ABSTRACT_OVITO_CLASS(ParaViewVTMFileFilter);
+IMPLEMENT_CREATABLE_OVITO_CLASS(ParaViewVTMImporter);
+OVITO_CLASSINFO(ParaViewVTMImporter, "DisplayName", "VTM");
 DEFINE_PROPERTY_FIELD(ParaViewVTMImporter, uniteMeshes);
 SET_PROPERTY_FIELD_LABEL(ParaViewVTMImporter, uniteMeshes, "Unite all meshes");
 
@@ -126,6 +127,8 @@ std::vector<ParaViewVTMBlockInfo> ParaViewVTMImporter::loadVTMFile(const FileHan
 
                 // Parse value of 'file' attribute if present.
                 QString file = xml.attributes().value("file").toString();
+                // Aspherix fix: The Windows version may include backslashes in the file path.
+                file.replace('\\', '/');
                 // Resolve file path and record the URL, which will be loaded later.
                 if(!file.isEmpty())
                     blockInfo.location = fileHandle.sourceUrl().resolved(QUrl(file));
@@ -148,8 +151,10 @@ std::vector<ParaViewVTMBlockInfo> ParaViewVTMImporter::loadVTMFile(const FileHan
                 // Also count the number of block pieces that are not empty.
                 int pieceCount = 0;
                 auto iter = std::find_if(datasetList.rbegin(), datasetList.rend(), [&](const ParaViewVTMBlockInfo& block) {
-                    if(block.blockPath != blockBranch) return true;
-                    if(!block.location.isEmpty()) pieceCount++;
+                    if(block.blockPath != blockBranch)
+                        return true;
+                    if(!block.location.isEmpty())
+                        pieceCount++;
                     return false;
                 });
                 OVITO_ASSERT(pieceCount <= datasetList.size());
@@ -157,8 +162,10 @@ std::vector<ParaViewVTMBlockInfo> ParaViewVTMImporter::loadVTMFile(const FileHan
                 int pieceIndex = pieceCount;
                 std::for_each(datasetList.rbegin(), iter, [&](ParaViewVTMBlockInfo& block) {
                     block.pieceCount = pieceCount;
-                    if(!block.location.isEmpty()) block.pieceIndex = --pieceIndex;
-                    else block.pieceIndex = -1;
+                    if(!block.location.isEmpty())
+                        block.pieceIndex = --pieceIndex;
+                    else
+                        block.pieceIndex = -1;
                 });
                 OVITO_ASSERT(pieceIndex == 0);
                 blockBranch.pop_back();
@@ -202,7 +209,7 @@ Future<PipelineFlowState> ParaViewVTMImporter::loadFrame(const LoadOperationRequ
     std::vector<ParaViewVTMBlockInfo> blockDatasets = loadVTMFile(request.fileHandle);
 
     // Create filter objects.
-    static const QVector<OvitoClassPtr> filterClassList = PluginManager::instance().listClasses(ParaViewVTMFileFilter::OOClass());
+    static const std::vector<OvitoClassPtr> filterClassList = PluginManager::instance().listClasses(ParaViewVTMFileFilter::OOClass());
     for(OvitoClassPtr clazz : filterClassList) {
         modifiedRequest.filters.push_back(static_object_cast<ParaViewVTMFileFilter>(clazz->createInstance()));
 
@@ -210,102 +217,105 @@ Future<PipelineFlowState> ParaViewVTMImporter::loadFrame(const LoadOperationRequ
         modifiedRequest.filters.back()->preprocessDatasets(blockDatasets, modifiedRequest, *this);
     }
 
-    OORef<PipelineNode> fileSource = request.pipelineNode.data();
-    if(!fileSource)
-        throw Exception(QStringLiteral("Object requesting the data import has been deleted."));
-
     // Load each dataset referenced by the VTM file.
-    Future<ExtendedLoadRequest> future = reduce_sequential(std::move(modifiedRequest), std::move(blockDatasets), ObjectExecutor(fileSource, true), [](const ParaViewVTMBlockInfo& blockInfo, ExtendedLoadRequest& request) {
+    Future<ExtendedLoadRequest> future = reduce_sequential(
+        std::move(modifiedRequest), std::move(blockDatasets), DeferredObjectExecutor(request.pipelineNode),
+        [](const ParaViewVTMBlockInfo& blockInfo, ExtendedLoadRequest& request) {
+            // We can skip empty datasets which are not associated with a VTK file.
+            if(blockInfo.location.isEmpty()) return Future<void>::createImmediateEmpty();
 
-        // We can skip empty datasets which are not associated with a VTK file.
-        if(blockInfo.location.isEmpty())
-            return Future<>::createImmediateEmpty();
+            // Set up the load request submitted to the FileSourceImporter.
+            request.dataBlockPrefix = blockInfo.blockPath.back();
+            request.blockInfo = blockInfo;
+            request.appendData =
+                (blockInfo.pieceIndex >
+                 0);  // Append data (instead of replacing it) when loading subsequent partial blocks of a piece-wise (parallel) dataset.
 
-        // Set up the load request submitted to the FileSourceImporter.
-        request.dataBlockPrefix = blockInfo.blockPath.back();
-        request.blockInfo = blockInfo;
-        request.appendData = (blockInfo.pieceIndex > 0); // Append data (instead of replacing it) when loading subsequent partial blocks of a piece-wise (parallel) dataset.
+            OORef<const PipelineNode> fileSource = request.pipelineNode.lock();
+            if(!fileSource) throw Exception(QStringLiteral("Object requesting the data import has been deleted."));
 
-        OORef<PipelineNode> fileSource = request.pipelineNode.data();
-        if(!fileSource)
-            throw Exception(QStringLiteral("Object requesting the data import has been deleted."));
+            // Retrieve the data file, then detect its format.
+            // Note: FileImporter::autodetectFileFormat() may only be called from the main thread.
+            return Application::instance()->fileManager()
+                .fetchUrl(blockInfo.location)
+                .then(ObjectExecutor(fileSource), [&request](const SharedFuture<FileHandle>& fileFuture) mutable -> Future<void> {
+                    OVITO_ASSERT(this_task::get());
 
-        // Retrieve the data file, then detect its format.
-        // Note: FileImporter::autodetectFileFormat() may only be called from the main thread.
-        return Application::instance()->fileManager().fetchUrl(blockInfo.location).then(*fileSource, [&request](const SharedFuture<FileHandle>& fileFuture) mutable -> Future<> {
-            OVITO_ASSERT(ExecutionContext::current().isValid());
-
-            try {
-                // Obtain a handle to the referenced data file.
-                const FileHandle& file = fileFuture.result();
-
-                // Give plugin filter objects the possibility to override the loading of the data file.
-                for(const auto& filter : request.filters) {
-                    Future<> future = filter->loadDataset(request.blockInfo, file, request);
-                    if(future.isValid())
-                        return future;
-                }
-                // If none of the filter objects decided to handle the loading process, fall back to our standard procedure,
-                // which consists of detecting the file's format and delegating the file parsing to the corresponding FileSourceImporter class.
-
-                // Detect file format and create an importer for it.
-                // This currently works only for FileSourceImporters. Files handled by other kinds of importers will be skipped.
-                // VTK dataset blocks using a file format not supported by OVITO are silently ignored.
-                OORef<FileSourceImporter> importer = dynamic_object_cast<FileSourceImporter>(FileImporter::autodetectFileFormat(file));
-                if(!importer)
-                    return Future<>::createImmediateEmpty();
-
-                // Remember the current status returned by the loading operations completed so far.
-                // We will prepend this existing status text to the one generated by the current file importer.
-                PipelineStatus lastStatus = request.state.status();
-
-                // Set up the load request submitted to the FileSourceImporter.
-                request.frame = Frame(file);
-                request.fileHandle = file;
-                request.state.setStatus(PipelineStatus::Success);
-
-                // Give plugin filter objects the possibility to pass additional information to the specific FileSourceImporter.
-                for(const auto& filter : request.filters)
-                    filter->configureImporter(request.blockInfo, request, importer);
-
-                // Parse the referenced file.
-                // Note: We need to keep the FileSourceImporter object while the asynchronous parsing process is
-                // in progress. That's why we store an otherwise unused pointer to it in the lambda function.
-                return importer->loadFrame(request).then([importer, filename = file.sourceUrl().fileName(), &request, lastStatus](Future<PipelineFlowState> blockDataFuture) mutable {
                     try {
-                        request.state = blockDataFuture.result();
+                        // Obtain a handle to the referenced data file.
+                        const FileHandle& file = fileFuture.result();
 
-                        // Concatenate status strings for the data blocks loaded so far.
-                        QString statusString = lastStatus.text();
-                        if(!request.state.status().text().isEmpty()) {
-                            // Append only the status text of the last partial dataset when loading a partial dataset.
-                            if(request.blockInfo.pieceIndex < 0 || request.blockInfo.pieceIndex == request.blockInfo.pieceCount - 1) {
-                                if(!statusString.isEmpty() && !statusString.endsWith(QChar('\n'))) statusString += QChar('\n');
-                                statusString += request.state.status().text();
-                            }
+                        // Give plugin filter objects the possibility to override the loading of the data file.
+                        for(const auto& filter : request.filters) {
+                            if(Future<void> future = filter->loadDataset(request.blockInfo, file, request))
+                                return future;
                         }
+                        // If none of the filter objects decided to handle the loading process, fall back to our standard procedure,
+                        // which consists of detecting the file's format and delegating the file parsing to the corresponding
+                        // FileSourceImporter class.
 
-                        // Also calculate a combined status code.
-                        PipelineStatus::StatusType statusType = lastStatus.type();
-                        if(statusType == PipelineStatus::Success || (statusType == PipelineStatus::Warning && request.state.status().type() == PipelineStatus::Error))
-                            statusType = request.state.status().type();
-                        request.state.setStatus(PipelineStatus(statusType, std::move(statusString)));
+                        // Detect file format and create an importer for it.
+                        // This currently works only for FileSourceImporters. Files handled by other kinds of importers will be skipped.
+                        // VTK dataset blocks using a file format not supported by OVITO are silently ignored.
+                        OORef<FileSourceImporter> importer = dynamic_object_cast<FileSourceImporter>(FileImporter::autodetectFileFormat(file));
+                        if(!importer)
+                            return Future<void>::createImmediateEmpty();
+
+                        // Remember the current status returned by the loading operations completed so far.
+                        // We will prepend this existing status text to the one generated by the current file importer.
+                        PipelineStatus lastStatus = request.state.status();
+
+                        // Set up the load request submitted to the FileSourceImporter.
+                        request.frame = Frame(file);
+                        request.fileHandle = file;
+                        request.state.setStatus(PipelineStatus::Success);
+
+                        // Give plugin filter objects the possibility to pass additional information to the specific FileSourceImporter.
+                        for(const auto& filter : request.filters) filter->configureImporter(request.blockInfo, request, importer);
+
+                        // Parse the referenced file.
+                        // Note: We need to keep the FileSourceImporter object while the asynchronous parsing process is
+                        // in progress. That's why we store an otherwise unused pointer to it in the lambda function.
+                        return importer->loadFrame(request).then([importer, filename = file.sourceUrl().fileName(), &request,
+                                                                  lastStatus](Future<PipelineFlowState> blockDataFuture) mutable {
+                            try {
+                                request.state = blockDataFuture.result();
+
+                                // Concatenate status strings for the data blocks loaded so far.
+                                QString statusString = lastStatus.text();
+                                if(!request.state.status().text().isEmpty()) {
+                                    // Append only the status text of the last partial dataset when loading a partial dataset.
+                                    if(request.blockInfo.pieceIndex < 0 ||
+                                       request.blockInfo.pieceIndex == request.blockInfo.pieceCount - 1) {
+                                        if(!statusString.isEmpty() && !statusString.endsWith(QChar('\n'))) statusString += QChar('\n');
+                                        statusString += request.state.status().text();
+                                    }
+                                }
+
+                                // Also calculate a combined status code.
+                                PipelineStatus::StatusType statusType = lastStatus.type();
+                                if(statusType == PipelineStatus::Success ||
+                                   (statusType == PipelineStatus::Warning && request.state.status().type() == PipelineStatus::Error))
+                                    statusType = request.state.status().type();
+                                request.state.setStatus(PipelineStatus(statusType, std::move(statusString)));
+                            }
+                            catch(Exception& ex) {
+                                throw ex.prependGeneralMessage(
+                                    tr("Failed to load VTK multi-block dataset '%1': %2").arg(request.dataBlockPrefix).arg(filename));
+                            }
+                        });
                     }
                     catch(Exception& ex) {
-                        throw ex.prependGeneralMessage(tr("Failed to load VTK multi-block dataset '%1': %2").arg(request.dataBlockPrefix).arg(filename));
+                        // Handle file errors, e.g. if the data block file referenced in the VTM file does not exist.
+                        request.state.setStatus(PipelineStatus(ex, QStringLiteral(" ")));
+                        ex.prependGeneralMessage(tr("Failed to access data file referenced by block '%1' in VTK multi-block file.")
+                                                     .arg(request.dataBlockPrefix));
+                        this_task::ui()->reportError(ex);
+                        // We treat such an error as recoverable and continue with loading the remaining data blocks.
+                        return Future<void>::createImmediateEmpty();
                     }
                 });
-            }
-            catch(Exception& ex) {
-                // Handle file errors, e.g. if the data block file referenced in the VTM file does not exist.
-                request.state.setStatus(PipelineStatus(ex, QStringLiteral(" ")));
-                ex.prependGeneralMessage(tr("Failed to access data file referenced by block '%1' in VTK multi-block file.").arg(request.dataBlockPrefix));
-                ExecutionContext::current().ui().reportError(ex);
-                // We treat such an error as recoverable and continue with loading the remaining data blocks.
-                return Future<>::createImmediateEmpty();
-            }
         });
-    });
 
     return future.then([](ExtendedLoadRequest&& request) -> PipelineFlowState {
 
@@ -325,7 +335,7 @@ void ParaViewVTMImporter::propertyChanged(const PropertyFieldDescriptor* field)
 {
     FileSourceImporter::propertyChanged(field);
 
-    if(field == PROPERTY_FIELD(uniteMeshes)) {
+    if(field == PROPERTY_FIELD(uniteMeshes) && !isBeingLoaded()) {
         // Reload input file(s) when this option is changed by the user.
         // There is no need to refetch the data file(s) from the remote location though. Reparsing the cached files is sufficient.
         requestReload();
