@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -28,8 +28,9 @@
 
 namespace Ovito {
 
-IMPLEMENT_OVITO_CLASS(ParaViewVTPMeshImporter);
-IMPLEMENT_OVITO_CLASS(MeshParaViewVTMFileFilter);
+IMPLEMENT_CREATABLE_OVITO_CLASS(ParaViewVTPMeshImporter);
+OVITO_CLASSINFO(ParaViewVTPMeshImporter, "DisplayName", "VTP");
+IMPLEMENT_CREATABLE_OVITO_CLASS(MeshParaViewVTMFileFilter);
 
 /******************************************************************************
 * Checks if the given file has format that can be read by this importer.
@@ -70,11 +71,13 @@ bool ParaViewVTPMeshImporter::OOMetaClass::checkFileFormat(const FileHandle& fil
 ******************************************************************************/
 void ParaViewVTPMeshImporter::FrameLoader::loadFile()
 {
-    setProgressText(tr("Reading ParaView VTP PolyData file %1").arg(fileHandle().toString()));
+    TaskProgress progress(this_task::ui());
+    progress.setText(tr("Reading ParaView VTP PolyData file %1").arg(fileHandle().toString()));
 
     // Create the destination mesh object.
     QString meshIdentifier = loadRequest().dataBlockPrefix;
-    if(meshIdentifier.isEmpty()) meshIdentifier = "mesh";
+    if(meshIdentifier.isEmpty())
+        meshIdentifier = "mesh";
     SurfaceMesh* mesh = state().getMutableLeafObject<SurfaceMesh>(SurfaceMesh::OOClass(), meshIdentifier);
     if(!mesh) {
         mesh = state().createObject<SurfaceMesh>(pipelineNode());
@@ -88,11 +91,13 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
         }
         if(!loadRequest().dataBlockPrefix.isEmpty()) {
             mesh->setTitle(tr("Mesh: %1").arg(loadRequest().dataBlockPrefix));
-            if(vis) vis->setTitle(tr("Mesh: %1").arg(loadRequest().dataBlockPrefix));
+            if(vis)
+                vis->setTitle(tr("Mesh: %1").arg(loadRequest().dataBlockPrefix));
         }
         else {
             mesh->setTitle(tr("Mesh"));
-            if(vis) vis->setTitle(tr("Mesh"));
+            if(vis)
+                vis->setTitle(tr("Mesh"));
         }
     }
     SurfaceMeshBuilder meshBuilder(mesh);
@@ -117,11 +122,11 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
     SurfaceMesh::face_index faceBaseIndex = SurfaceMesh::InvalidIndex;
     std::vector<PropertyPtr> cellDataArrays;
     std::vector<PropertyPtr> pointDataArrays;
+    int vtkHeaderType = 8; // Assume UInt64 by default
 
     // Parse the elements of the XML file.
     while(xml.readNextStartElement()) {
-        if(isCanceled())
-            return;
+        this_task::throwIfCanceled();
 
         if(xml.name().compare(QStringLiteral("VTKFile")) == 0) {
             if(xml.attributes().value("type").compare(QStringLiteral("PolyData")) != 0)
@@ -130,6 +135,8 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
                 xml.raiseError(tr("Byte order must be 'LittleEndian'. Please ask the OVITO developers to extend the capabilities of the file parser."));
             else if(!xml.attributes().value("compressor").isEmpty())
                 xml.raiseError(tr("Current implementation does not support compressed data arrays. Please ask the OVITO developers to extend the capabilities of the file parser."));
+            if(xml.attributes().value("header_type").compare(QStringLiteral("UInt32")) == 0)
+                vtkHeaderType = 4;
         }
         else if(xml.name().compare(QStringLiteral("PolyData")) == 0) {
             // Do nothing. Parse child elements.
@@ -150,7 +157,7 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
             // Parse child <DataArray> element containing the point coordinates.
             if(!xml.readNextStartElement())
                 break;
-            PropertyPtr property = parseDataArray(xml, Property::FloatDefault);
+            PropertyPtr property = parseDataArray(xml, vtkHeaderType, Property::FloatDefault);
             if(!property)
                 break;
 
@@ -168,7 +175,7 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
             // Parse child <DataArray> element containing the connectivity information.
             if(!xml.readNextStartElement())
                 break;
-            PropertyPtr connectivityArray = parseDataArray(xml, DataBufferPrimitiveType<SurfaceMesh::vertex_index>::value);
+            PropertyPtr connectivityArray = parseDataArray(xml, vtkHeaderType, DataBufferPrimitiveType<SurfaceMesh::vertex_index>::value);
             if(!connectivityArray)
                 break;
             // Make sure the data array has the expected data layout.
@@ -181,7 +188,7 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
             // Parse child <DataArray> element containing the offset information.
             if(!xml.readNextStartElement())
                 break;
-            PropertyPtr offsetsArray = parseDataArray(xml, Property::Int32);
+            PropertyPtr offsetsArray = parseDataArray(xml, vtkHeaderType, Property::Int32);
             if(!offsetsArray)
                 break;
             // Make sure the data array has the expected data layout.
@@ -213,9 +220,9 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
         }
         else if(xml.name().compare(QStringLiteral("CellData")) == 0) {
             // Parse <DataArray> child elements.
-            while(xml.readNextStartElement() && !isCanceled()) {
+            while(xml.readNextStartElement() && !this_task::isCanceled()) {
                 if(xml.name().compare(QStringLiteral("DataArray")) == 0) {
-                    if(PropertyPtr property = parseDataArray(xml))
+                    if(PropertyPtr property = parseDataArray(xml, vtkHeaderType))
                         cellDataArrays.push_back(std::move(property));
                     else
                         break;
@@ -227,9 +234,9 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
         }
         else if(xml.name().compare(QStringLiteral("PointData")) == 0) {
             // Parse child elements.
-            while(xml.readNextStartElement() && !isCanceled()) {
+            while(xml.readNextStartElement() && !this_task::isCanceled()) {
                 if(xml.name().compare(QStringLiteral("DataArray")) == 0) {
-                    if(PropertyPtr property = parseDataArray(xml))
+                    if(PropertyPtr property = parseDataArray(xml, vtkHeaderType))
                         pointDataArrays.push_back(std::move(property));
                     else
                         break;
@@ -253,8 +260,7 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
         throw Exception(tr("VTP file parsing error on line %1, column %2: %3")
             .arg(xml.lineNumber()).arg(xml.columnNumber()).arg(xml.errorString()));
     }
-    if(isCanceled())
-        return;
+    this_task::throwIfCanceled();
 
     // Add cell data arrays to the mesh.
     if(numberOfPolys == numberOfCells) {
@@ -268,13 +274,21 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
             if(!loadRequest().appendData) {
                 OVITO_ASSERT(property->size() == meshBuilder.faceCount());
                 OVITO_ASSERT(faceBaseIndex == 0);
-                meshBuilder.addFaceProperty(std::move(property));
+                const Property* existingProperty = property->isStandardProperty()
+                    ? meshBuilder.faceProperty(static_cast<SurfaceMeshFaces::Type>(property->typeId()))
+                    : meshBuilder.faceProperty(property->name());
+                if(!existingProperty)
+                    meshBuilder.addFaceProperty(std::move(property));
+                else
+                    meshBuilder.mutableFaces()->replaceReferencesTo(existingProperty, property);
             }
             else {
-                Property* existingProperty = property->type() != SurfaceMeshFaces::UserProperty
-                    ? meshBuilder.mutableFaceProperty(static_cast<SurfaceMeshFaces::Type>(property->type()))
+                Property* existingProperty = property->isStandardProperty()
+                    ? meshBuilder.mutableFaceProperty(static_cast<SurfaceMeshFaces::Type>(property->typeId()))
                     : meshBuilder.mutableFaceProperty(property->name());
-                if(existingProperty && existingProperty->dataType() == property->dataType() && existingProperty->componentCount() == property->componentCount()) {
+                if(existingProperty) {
+                    if(existingProperty->dataType() != property->dataType() || existingProperty->componentCount() != property->componentCount())
+                        throw Exception(tr("Per mesh face data array '%1' has incompatible data type or component count in some trajectory frame(s).").arg(property->name()));
                     existingProperty->copyRangeFrom(*property, 0, faceBaseIndex, property->size());
                 }
             }
@@ -292,13 +306,21 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
         if(!loadRequest().appendData) {
             OVITO_ASSERT(property->size() == meshBuilder.vertexCount());
             OVITO_ASSERT(vertexBaseIndex == 0);
-            meshBuilder.addVertexProperty(std::move(property));
+            const Property* existingProperty = property->isStandardProperty()
+                ? meshBuilder.vertexProperty(static_cast<SurfaceMeshVertices::Type>(property->typeId()))
+                : meshBuilder.vertexProperty(property->name());
+            if(!existingProperty)
+                meshBuilder.addVertexProperty(std::move(property));
+            else
+                meshBuilder.mutableVertices()->replaceReferencesTo(existingProperty, property);
         }
         else {
-            Property* existingProperty = property->type() != SurfaceMeshVertices::UserProperty
-                ? meshBuilder.mutableVertexProperty(static_cast<SurfaceMeshVertices::Type>(property->type()))
+            Property* existingProperty = property->isStandardProperty()
+                ? meshBuilder.mutableVertexProperty(static_cast<SurfaceMeshVertices::Type>(property->typeId()))
                 : meshBuilder.mutableVertexProperty(property->name());
-            if(existingProperty && existingProperty->dataType() == property->dataType() && existingProperty->componentCount() == property->componentCount()) {
+            if(existingProperty) {
+                if(existingProperty->dataType() != property->dataType() || existingProperty->componentCount() != property->componentCount())
+                    throw Exception(tr("Per mesh vertex data array '%1' has incompatible data type or component count in some trajectory frame(s).").arg(property->name()));
                 existingProperty->copyRangeFrom(*property, 0, vertexBaseIndex, property->size());
             }
         }
@@ -318,6 +340,9 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
             .arg(meshBuilder.vertexCount())
             .arg(meshBuilder.faceCount()));
     }
+#ifdef OVITO_DEBUG
+    meshBuilder.mesh()->verifyMeshIntegrity();
+#endif
 
     // Call base implementation.
     StandardFrameLoader::loadFile();
@@ -326,7 +351,7 @@ void ParaViewVTPMeshImporter::FrameLoader::loadFile()
 /******************************************************************************
 * Reads a <DataArray> element and returns it as an OVITO property.
 ******************************************************************************/
-PropertyPtr ParaViewVTPMeshImporter::FrameLoader::parseDataArray(QXmlStreamReader& xml, int convertToDataType)
+PropertyPtr ParaViewVTPMeshImporter::FrameLoader::parseDataArray(QXmlStreamReader& xml, int vtkHeaderType, int convertToDataType)
 {
     // Make sure it is really an <DataArray>.
     if(xml.name().compare(QStringLiteral("DataArray")) != 0) {
@@ -336,18 +361,11 @@ PropertyPtr ParaViewVTPMeshImporter::FrameLoader::parseDataArray(QXmlStreamReade
 
     // Check value of the 'format' attribute.
     QString format = xml.attributes().value("format").toString();
-    bool isBinary;
     if(format.isEmpty()) {
         xml.raiseError(tr("Expected 'format' attribute in <%1> element.").arg(xml.name().toString()));
         return {};
     }
-    else if(format == "binary") {
-        isBinary = true;
-    }
-    else if(format == "ascii") {
-        isBinary = false;
-    }
-    else {
+    else if(format != "binary" && format != "ascii") {
         xml.raiseError(tr("Invalid value of 'format' attribute in <%1> element: %2").arg(xml.name().toString()).arg(format));
         return {};
     }
@@ -369,11 +387,22 @@ PropertyPtr ParaViewVTPMeshImporter::FrameLoader::parseDataArray(QXmlStreamReade
         else convertToDataType = Property::FloatDefault;
     }
 
+    // Parse optional list of vector component names.
+    QStringList componentNames;
+    for(int c = 0; c < numComponents; ++c) {
+        QString componentName = xml.attributes().value(QStringLiteral("ComponentName%1").arg(c)).toString();
+        if(componentName.isEmpty()) {
+            componentNames.clear();
+            break;
+        }
+        componentNames.push_back(Property::makeComponentNameValid(componentName));
+    }
+
     // Create destination property. Initially with zero elements, will be resized later when the size of the VTK data array is known.
-    PropertyPtr property = DataOORef<Property>::create(DataBuffer::Uninitialized, 0, convertToDataType, numComponents, Property::makePropertyNameValid(name));
+    PropertyPtr property = DataOORef<Property>::create(DataBuffer::Uninitialized, 0, convertToDataType, numComponents, Property::makePropertyNameValid(name), 0, std::move(componentNames));
 
     // Delegate parsing of payload to sub-routine.
-    if(!parseVTKDataArray(property.get(), xml))
+    if(!parseVTKDataArray(property.get(), vtkHeaderType, xml))
         return {};
 
     return property;
@@ -386,7 +415,6 @@ static inline void tokenizeString(const QString& str, F&& f)
     QStringView textView(str);
     auto start = textView.cbegin();
     auto eos = textView.cend();
-    size_t valueCount = 0;
     while(start != eos) {
         // Skip whitespace characters.
         while(start != eos && start->isSpace())
@@ -406,7 +434,7 @@ static inline void tokenizeString(const QString& str, F&& f)
 /******************************************************************************
 * Reads a <DataArray> element and stores it in the given OVITO data buffer.
 ******************************************************************************/
-bool ParaViewVTPMeshImporter::parseVTKDataArray(DataBuffer* buffer, QXmlStreamReader& xml, int vectorComponent, size_t destBaseIndex)
+bool ParaViewVTPMeshImporter::parseVTKDataArray(DataBuffer* buffer, int vtkHeaderType, QXmlStreamReader& xml, int vectorComponent, size_t destBaseIndex, size_t replication)
 {
     // Make sure it is really an <DataArray>.
     if(xml.name().compare(QStringLiteral("DataArray")) != 0) {
@@ -480,10 +508,13 @@ bool ParaViewVTPMeshImporter::parseVTKDataArray(DataBuffer* buffer, QXmlStreamRe
         byteArray = QByteArray::fromBase64(text.toLatin1());
 
         // Note: Decoded binary data is prepended with array size information.
-        qint64 dataArraySize = (byteArray.size() >= sizeof(qint64)) ? *reinterpret_cast<const qint64*>(byteArray.constData()) : -1;
-        if(dataArraySize + sizeof(qint64) != byteArray.size()) {
+        qint64 dataArraySize =
+            vtkHeaderType == 8 ?
+                ((byteArray.size() >= sizeof(qint64)) ? *reinterpret_cast<const qint64*>(byteArray.constData()) : -1) :
+                ((byteArray.size() >= sizeof(qint32)) ? *reinterpret_cast<const qint32*>(byteArray.constData()) : -1);
+        if(dataArraySize + vtkHeaderType != byteArray.size()) {
             xml.raiseError(tr("Data array size mismatch: Expected %1 bytes of base64 encoded data, but XML element contains %2 bytes.")
-                .arg(dataArraySize + sizeof(qint64))
+                .arg(dataArraySize + vtkHeaderType)
                 .arg(byteArray.size()));
             return false;
         }
@@ -495,7 +526,7 @@ bool ParaViewVTPMeshImporter::parseVTKDataArray(DataBuffer* buffer, QXmlStreamRe
             return false;
         }
 
-        rawDataPtr = byteArray.constData() + sizeof(qint64);
+        rawDataPtr = byteArray.constData() + vtkHeaderType;
     }
     else {
         // Tokenize the XML element contents.
@@ -550,8 +581,8 @@ bool ParaViewVTPMeshImporter::parseVTKDataArray(DataBuffer* buffer, QXmlStreamRe
     }
 
     // Check if VTK data array size fits to the size of the target buffer provided by the caller.
-    if(buffer->size() != 0 && buffer->size() != elementCount + destBaseIndex) {
-        xml.raiseError(tr("Data array size mismatch: Expected %1 data tuples, but <DataArray> element contains %2 tuples.").arg(buffer->size() - destBaseIndex).arg(elementCount));
+    if(buffer->size() != 0 && buffer->size() != elementCount * replication + destBaseIndex) {
+        xml.raiseError(tr("Data array size mismatch: Expected %1 data tuples, but <DataArray> element contains %2 tuples.").arg((buffer->size() - destBaseIndex) / replication).arg(elementCount));
         return false;
     }
     if(vectorComponent == -1) {
@@ -570,11 +601,11 @@ bool ParaViewVTPMeshImporter::parseVTKDataArray(DataBuffer* buffer, QXmlStreamRe
     // Allocate destination buffer (if not already done).
     if(buffer->size() == 0) {
         OVITO_ASSERT(destBaseIndex == 0);
-        buffer->resize(elementCount, false);
+        buffer->resize(elementCount * replication, false);
     }
 
     // Verify parameters.
-    if(destBaseIndex + elementCount > buffer->size()) {
+    if(destBaseIndex + elementCount * replication > buffer->size()) {
         xml.raiseError(tr("Data array size mismatch: Number of elements in the <DataArray> exceeds expected range."));
         return false;
     }
@@ -582,14 +613,38 @@ bool ParaViewVTPMeshImporter::parseVTKDataArray(DataBuffer* buffer, QXmlStreamRe
     auto copyValuesToBuffer = [&](auto srcData) {
         const auto begin = srcData;
         const auto end = begin + elementCount * numComponents;
-        bool discardExistingData = (destBaseIndex == 0);
+        DataBuffer::BufferInitialization existingDataAccessMode = (destBaseIndex == 0) ? DataBuffer::Uninitialized : DataBuffer::Initialized;
 
         buffer->forAnyType([&](auto _) {
             using T = decltype(_);
-            if(vectorComponent == -1)
-                std::copy(begin, end, std::next(BufferWriteAccess<T*, access_mode::write>(buffer, discardExistingData).begin(), destBaseIndex * buffer->componentCount()));
-            else
-                std::copy(begin, end, std::next(std::begin(BufferWriteAccess<T*, access_mode::write>(buffer, discardExistingData).componentRange(vectorComponent)), destBaseIndex));
+            if(replication == 1) {
+                if(vectorComponent == -1)
+                    std::copy(begin, end, std::next(BufferWriteAccess<T*, access_mode::write>(buffer, existingDataAccessMode).begin(), destBaseIndex * buffer->componentCount()));
+                else
+                    std::copy(begin, end, std::next(std::begin(BufferWriteAccess<T*, access_mode::write>(buffer, existingDataAccessMode).componentRange(vectorComponent)), destBaseIndex));
+            }
+            else {
+                BufferWriteAccess<T*, access_mode::write> dstAccess(buffer, existingDataAccessMode);
+                if(vectorComponent == -1) {
+                    auto dst = std::next(dstAccess.begin(), destBaseIndex * buffer->componentCount());
+                    for(auto i = begin; i != end; i += numComponents) {
+                        for(size_t j = 0; j < replication; ++j) {
+                            auto i2 = i;
+                            for(size_t c = 0; c < numComponents; ++c, ++i2) {
+                                *dst++ = *i2;
+                            }
+                        }
+                    }
+                    OVITO_ASSERT(dst == dstAccess.end());
+                }
+                else {
+                    auto dst = std::next(std::begin(BufferWriteAccess<T*, access_mode::write>(buffer, existingDataAccessMode).componentRange(vectorComponent)), destBaseIndex);
+                    for(auto i = begin; i != end; ++i) {
+                        for(size_t j = 0; j < replication; ++j)
+                            *dst++ = *i;
+                    }
+                }
+            }
         });
     };
 
@@ -634,7 +689,7 @@ bool ParaViewVTPMeshImporter::parseVTKDataArray(DataBuffer* buffer, QXmlStreamRe
 void MeshParaViewVTMFileFilter::preprocessDatasets(std::vector<ParaViewVTMBlockInfo>& blockDatasets, FileSourceImporter::LoadOperationRequest& request, const ParaViewVTMImporter& vtmImporter)
 {
     // Special handling of meshes that are grouped in the "Meshes" block of an Aspherix VTM file.
-    // This is specific bheavior for VTM files written by the Aspherix code.
+    // This is specific behavior for VTM files written by the Aspherix code.
     if(vtmImporter.uniteMeshes()) {
         // Count the total number of mesh data files referenced in the "Meshes" sections of the VTM file.
         int numMeshFiles = boost::count_if(blockDatasets, [](const ParaViewVTMBlockInfo& block) {
@@ -681,6 +736,21 @@ void MeshParaViewVTMFileFilter::preprocessDatasets(std::vector<ParaViewVTMBlockI
         ConstDataObjectPath path = request.state.getObject<SurfaceMesh>(QStringLiteral("combined"));
         if(path.size() == 1)
             request.state.mutableData()->removeObject(path.back());
+
+        // Remove meshes from those trajectory frames where the VTM file does not list them.
+        // But instead of deleting the mesh objects from the data collection just clear their contents,
+        // such that visual element settings are preserved across frames even if a mesh temporarily disappears.
+        for(qsizetype i = 0; i < request.state.data()->objects().size(); i++) {
+            if(const SurfaceMesh* mesh = dynamic_object_cast<SurfaceMesh>(request.state.data()->objects()[i])) {
+                if(mesh->vertices()->elementCount() != 0 || mesh->faces()->elementCount() != 0) {
+                    SurfaceMesh* mutableMesh = request.state.mutableData()->makeMutable(mesh);
+                    mutableMesh->makeTopologyMutable()->clear();
+                    mutableMesh->makeVerticesMutable()->setElementCount(0);
+                    mutableMesh->makeFacesMutable()->setElementCount(0);
+                    mutableMesh->makeRegionsMutable()->setElementCount(0);
+                }
+            }
+        }
     }
 }
 
