@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -34,17 +34,6 @@ SurfaceMeshBuilder::SurfaceMeshBuilder(SurfaceMesh* mesh) : SurfaceMeshReadAcces
     OVITO_ASSERT(!mesh || mesh->isSafeToModify());
 }
 
-#ifdef OVITO_DEBUG
-/******************************************************************************
-* Destructor.
-******************************************************************************/
-SurfaceMeshBuilder::~SurfaceMeshBuilder()
-{
-    if(mesh())
-        mesh()->verifyMeshIntegrity();
-}
-#endif
-
 /******************************************************************************
 * Resets the surface mesh structure by discarding all existing vertices, faces and regions.
 ******************************************************************************/
@@ -335,8 +324,11 @@ void SurfaceMeshBuilder::deleteIsolatedVertices()
 /******************************************************************************
 * Fairs a closed triangle mesh.
 ******************************************************************************/
-bool SurfaceMeshBuilder::smoothMesh(int numIterations, ProgressingTask& task, FloatType k_PB, FloatType lambda)
+void SurfaceMeshBuilder::smoothMesh(int numIterations, TaskProgress& progress, FloatType k_PB, FloatType lambda)
 {
+    if(numIterations <= 0)
+        return;
+
     // This is the implementation of the mesh smoothing algorithm:
     //
     // Gabriel Taubin
@@ -344,13 +336,13 @@ bool SurfaceMeshBuilder::smoothMesh(int numIterations, ProgressingTask& task, Fl
     // In SIGGRAPH 95 Conference Proceedings, pages 351-358 (1995)
 
     // Performs one iteration of the smoothing algorithm.
-    auto smoothMeshIteration = [this](FloatType prefactor) {
+    auto smoothMeshIteration = [&](FloatType prefactor) {
 
         BufferReadAccess<Point3> vertexPositions(expectVertexProperty(SurfaceMeshVertices::PositionProperty));
 
         // Compute displacement for each vertex.
         std::vector<Vector3> displacements(vertexCount());
-        parallelFor(vertexCount(), [&](vertex_index vertex) {
+        parallelFor(vertexCount(), 4096, progress, [&](vertex_index vertex) {
             Vector3 d = Vector3::Zero();
 
             // Go in positive direction around vertex, facet by facet.
@@ -379,16 +371,15 @@ bool SurfaceMeshBuilder::smoothMesh(int numIterations, ProgressingTask& task, Fl
     };
 
     FloatType mu = FloatType(1) / (k_PB - FloatType(1)/lambda);
-    task.setProgressMaximum(numIterations);
 
+    progress.beginSubSteps(2 * numIterations);
     for(int iteration = 0; iteration < numIterations; iteration++) {
-        if(!task.setProgressValue(iteration))
-            return false;
         smoothMeshIteration(lambda);
+        progress.nextSubStep();
         smoothMeshIteration(mu);
+        progress.nextSubStep();
     }
-
-    return !task.isCanceled();
+    progress.endSubSteps();
 }
 
 /******************************************************************************
@@ -801,20 +792,20 @@ SurfaceMeshBuilder::AggregateVolumes SurfaceMeshBuilder::computeAggregateVolumes
 }
 
 /******************************************************************************
- * Set the volume of external regions to infinity if the simulation cell is non-periodic.
+ * Set the volume of the external region to infinity if the simulation cell
+ * is non-periodic in at least one spatial direction.
  ******************************************************************************/
-void SurfaceMeshBuilder::nonPBCexternalVolume()
+void SurfaceMeshBuilder::setExternalRegionVolumeInfinityIfNonPeriodic()
 {
-    // PBC is periodic. Nothing to do.
-    if(domain()->pbcX() && domain()->pbcY() && domain()->pbcZ()) {
+    // If the domain is 3d periodic, do nothing.
+    if(domain()->pbcX() && domain()->pbcY() && domain()->pbcZ())
         return;
-    }
 
     BufferReadAccess<SelectionIntType> isFilled = expectRegionProperty(SurfaceMeshRegions::IsFilledProperty);
     BufferReadAccess<SelectionIntType> isExterior = expectRegionProperty(SurfaceMeshRegions::IsExteriorProperty);
     BufferWriteAccess<FloatType, access_mode::write> regionVolumes = mutableRegionProperty(SurfaceMeshRegions::VolumeProperty);
 
-    for(SurfaceMesh::region_index region{0}; region < regionCount(); region++) {
+    for(SurfaceMesh::region_index region = 0; region < regionCount(); region++) {
         if(!isFilled[region] && isExterior[region]) {
             regionVolumes[region] = std::numeric_limits<FloatType>::infinity();
         }
