@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -27,7 +27,8 @@
 
 namespace Ovito {
 
-IMPLEMENT_OVITO_CLASS(SurfaceMeshTopology);
+IMPLEMENT_CREATABLE_OVITO_CLASS(SurfaceMeshTopology);
+OVITO_CLASSINFO(SurfaceMeshTopology, "DisplayName", "Surface mesh topology");
 
 constexpr SurfaceMeshTopology::size_type SurfaceMeshTopology::InvalidIndex;
 
