@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -33,7 +33,8 @@
 
 namespace Ovito {
 
-IMPLEMENT_OVITO_CLASS(LAMMPSBinaryDumpImporter);
+IMPLEMENT_CREATABLE_OVITO_CLASS(LAMMPSBinaryDumpImporter);
+OVITO_CLASSINFO(LAMMPSBinaryDumpImporter, "DisplayName", "LAMMPS Binary Dump");
 DEFINE_PROPERTY_FIELD(LAMMPSBinaryDumpImporter, columnMapping);
 SET_PROPERTY_FIELD_LABEL(LAMMPSBinaryDumpImporter, columnMapping, "File column mapping");
 
@@ -152,18 +153,19 @@ bool LAMMPSBinaryDumpImporter::OOMetaClass::checkFileFormat(const FileHandle& fi
 /******************************************************************************
 * Scans the data file and builds a list of source frames.
 ******************************************************************************/
-void LAMMPSBinaryDumpImporter::FrameFinder::discoverFramesInFile(QVector<FileSourceImporter::Frame>& frames)
+void LAMMPSBinaryDumpImporter::discoverFramesInFile(const FileHandle& fileHandle, QVector<FileSourceImporter::Frame>& frames) const
 {
     // Open input file in binary mode for reading.
-    std::unique_ptr<QIODevice> file = fileHandle().createIODevice();
+    std::unique_ptr<QIODevice> file = fileHandle.createIODevice();
     if(!file->open(QIODevice::ReadOnly))
         throw Exception(tr("Failed to open binary LAMMPS dump file: %1.").arg(file->errorString()));
 
-    setProgressText(tr("Scanning binary LAMMPS dump file %1").arg(fileHandle().toString()));
-    setProgressMaximum(file->size());
+    TaskProgress progress(this_task::ui());
+    progress.setText(tr("Scanning binary LAMMPS dump file %1").arg(fileHandle.toString()));
+    progress.setMaximum(file->size());
 
-    Frame frame(fileHandle());
-    while(!file->atEnd() && !isCanceled()) {
+    Frame frame(fileHandle);
+    while(!file->atEnd() && !this_task::isCanceled()) {
         frame.byteOffset = file->pos();
 
         // Parse file header.
@@ -185,8 +187,8 @@ void LAMMPSBinaryDumpImporter::FrameFinder::discoverFramesInFile(QVector<FileSou
             if(!file->seek(filePos))
                 throw Exception(tr("Unexpected end of file."));
 
-            if(!setProgressValue(filePos))
-                return;
+            // Update progress bar and check for user cancellation.
+            progress.setValue(filePos);
         }
 
         // Create a new record for the timestep.
@@ -346,7 +348,8 @@ bool LAMMPSBinaryDumpHeader::parse(QIODevice& input)
 ******************************************************************************/
 void LAMMPSBinaryDumpImporter::FrameLoader::loadFile()
 {
-    setProgressText(tr("Reading binary LAMMPS dump file %1").arg(fileHandle().toString()));
+    TaskProgress progress(this_task::ui());
+    progress.setText(tr("Reading binary LAMMPS dump file %1").arg(fileHandle().toString()));
 
     // Open input file for reading.
     std::unique_ptr<QIODevice> file = fileHandle().createIODevice();
@@ -366,7 +369,7 @@ void LAMMPSBinaryDumpImporter::FrameLoader::loadFile()
     if(header.simulationTime != std::numeric_limits<double>::lowest())
         state().setAttribute(QStringLiteral("Time"), QVariant::fromValue(header.simulationTime), pipelineNode());
 
-    setProgressMaximum(header.natoms);
+    progress.setMaximum(header.natoms);
     setParticleCount(header.natoms);
 
     // LAMMPS only stores the outer bounding box dimensions of the simulation cell in the dump file.
@@ -425,8 +428,8 @@ void LAMMPSBinaryDumpImporter::FrameLoader::loadFile()
             const double* iter = chunkData.constData();
             for(int nChunkAtoms = n / header.size_one; nChunkAtoms--; ++i, iter += header.size_one) {
 
-                // Update progress indicator.
-                if(!setProgressValueIntermittent(i)) return;
+                // Update progress bar and check for user cancellation.
+                progress.setValueIntermittent(i);
 
                 try {
                     columnParser.readElement(i, iter, header.size_one);
@@ -452,7 +455,7 @@ void LAMMPSBinaryDumpImporter::FrameLoader::loadFile()
         // the type of particle coordinates. Reduced coordinates are found in columns
         // "xs, ys, zs" or "xsu, ysu, zsu".
         for(int i = 0; i < (int)_columnMapping.size() && i < fileColumnNames.size(); i++) {
-            if(_columnMapping[i].property.type() == Particles::PositionProperty) {
+            if(_columnMapping[i].property.isStandardProperty(&Particles::OOClass(), Particles::PositionProperty)) {
                 reducedCoordinates = (
                         fileColumnNames[i] == "xs" || fileColumnNames[i] == "xsu" ||
                         fileColumnNames[i] == "ys" || fileColumnNames[i] == "ysu" ||
@@ -491,7 +494,7 @@ void LAMMPSBinaryDumpImporter::FrameLoader::loadFile()
         // If a "diameter" column was loaded and stored in the "Radius" particle property,
         // we need to divide values by two.
         for(int i = 0; i < (int)_columnMapping.size() && i < fileColumnNames.size(); i++) {
-            if(_columnMapping[i].property.type() == Particles::RadiusProperty && fileColumnNames[i] == "diameter") {
+            if(_columnMapping[i].property.isStandardProperty(&Particles::OOClass(), Particles::RadiusProperty) && fileColumnNames[i] == "diameter") {
                 if(BufferWriteAccess<GraphicsFloatType, access_mode::read_write> radiusProperty = particles()->getMutableProperty(Particles::RadiusProperty)) {
                     for(auto& r : radiusProperty)
                         r *= 0.5f;
@@ -502,7 +505,7 @@ void LAMMPSBinaryDumpImporter::FrameLoader::loadFile()
 
         // Same for the "c_diameter[1..3]" columns being mapped to the "Aspherical Shape" property.
         for(int i = 0; i < (int)_columnMapping.size() && i < fileColumnNames.size(); i++) {
-            if(_columnMapping[i].property.type() == Particles::AsphericalShapeProperty && (fileColumnNames[i] == "c_diameter[1]" || fileColumnNames[i] == "c_diameter[2]" || fileColumnNames[i] == "c_diameter[3]")) {
+            if(_columnMapping[i].property.isStandardProperty(&Particles::OOClass(), Particles::AsphericalShapeProperty) && (fileColumnNames[i] == "c_diameter[1]" || fileColumnNames[i] == "c_diameter[2]" || fileColumnNames[i] == "c_diameter[3]")) {
                 if(BufferWriteAccess<Vector3G, access_mode::read_write> shapeProperty = particles()->getMutableProperty(Particles::AsphericalShapeProperty)) {
                     for(auto& s : shapeProperty) {
                         s.x() *= 0.5f;
@@ -517,9 +520,9 @@ void LAMMPSBinaryDumpImporter::FrameLoader::loadFile()
 
     // Detect dimensionality of system. It's a 2D system if no file column has been mapped to the Position.Z particle property.
     if(std::none_of(_columnMapping.begin(), _columnMapping.end(), [](const InputColumnInfo& column) {
-        return column.property.type() == Particles::PositionProperty && column.property.vectorComponent() == 2;
+        return column.property.isStandardProperty(&Particles::OOClass(), Particles::PositionProperty) && column.property.componentIndex(&Particles::OOClass()) == 2;
     }) && std::any_of(_columnMapping.begin(), _columnMapping.end(), [](const InputColumnInfo& column) {
-        return column.property.type() == Particles::PositionProperty && column.property.vectorComponent() != 2;
+        return column.property.isStandardProperty(&Particles::OOClass(), Particles::PositionProperty) && column.property.componentIndex(&Particles::OOClass()) != 2;
     })) {
         simulationCell()->setIs2D(true);
     }
@@ -571,6 +574,18 @@ Future<ParticleInputColumnMapping> LAMMPSBinaryDumpImporter::inspectFileHeader(c
         });
 }
 
+/******************************************************************************
+* Is called when the value of a non-animatable property field of this RefMaker has changed.
+******************************************************************************/
+void LAMMPSBinaryDumpImporter::propertyChanged(const PropertyFieldDescriptor* field)
+{
+    ParticleImporter::propertyChanged(field);
+
+    if(field == PROPERTY_FIELD(columnMapping) && !isBeingLoaded()) {
+        requestReload();
+    }
+}
+
 /******************************************************************************
  * Saves the class' contents to the given stream.
  *****************************************************************************/
