@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -34,7 +34,10 @@
 
 namespace Ovito {
 
-IMPLEMENT_OVITO_CLASS(VoronoiAnalysisModifier);
+IMPLEMENT_CREATABLE_OVITO_CLASS(VoronoiAnalysisModifier);
+OVITO_CLASSINFO(VoronoiAnalysisModifier, "DisplayName", "Voronoi analysis");
+OVITO_CLASSINFO(VoronoiAnalysisModifier, "Description", "Determine nearest particle neighbors, atomic volume and Voronoi indices.");
+OVITO_CLASSINFO(VoronoiAnalysisModifier, "ModifierCategory", "Analysis");
 DEFINE_PROPERTY_FIELD(VoronoiAnalysisModifier, onlySelected);
 DEFINE_PROPERTY_FIELD(VoronoiAnalysisModifier, useRadii);
 DEFINE_PROPERTY_FIELD(VoronoiAnalysisModifier, computeIndices);
@@ -46,10 +49,10 @@ DEFINE_PROPERTY_FIELD(VoronoiAnalysisModifier, relativeFaceThreshold);
 DEFINE_REFERENCE_FIELD(VoronoiAnalysisModifier, bondsVis);
 DEFINE_REFERENCE_FIELD(VoronoiAnalysisModifier, polyhedraVis);
 SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, onlySelected, "Use only selected particles");
-SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, useRadii, "Use particle radii");
-SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, computeIndices, "Compute Voronoi indices");
-SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, computeBonds, "Generate neighbor bonds");
-SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, computePolyhedra, "Generate Voronoi polyhedra");
+SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, useRadii, "Use particle radii (polydisperse Voronoi tessellation)");
+SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, computeIndices, "Voronoi indices");
+SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, computeBonds, "Neighbor bonds");
+SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, computePolyhedra, "Voronoi polyhedra");
 SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, edgeThreshold, "Edge length threshold");
 SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, faceThreshold, "Absolute face area threshold");
 SET_PROPERTY_FIELD_LABEL(VoronoiAnalysisModifier, relativeFaceThreshold, "Relative face area threshold");
@@ -93,18 +96,12 @@ private:
 #endif
 
 /******************************************************************************
-* Constructs the modifier object.
+* Constructor.
 ******************************************************************************/
-VoronoiAnalysisModifier::VoronoiAnalysisModifier(ObjectInitializationFlags flags) : AsynchronousModifier(flags),
-    _onlySelected(false),
-    _useRadii(false),
-    _edgeThreshold(0),
-    _faceThreshold(0),
-    _computeIndices(false),
-    _computeBonds(false),
-    _computePolyhedra(false),
-    _relativeFaceThreshold(0)
+void VoronoiAnalysisModifier::initializeObject(ObjectInitializationFlags flags)
 {
+    Modifier::initializeObject(flags);
+
     if(!flags.testFlag(ObjectInitializationFlag::DontInitializeObject)) {
         // Create the vis element for rendering the bonds generated by the modifier.
         setBondsVis(OORef<BondsVis>::create(flags));
@@ -128,17 +125,53 @@ bool VoronoiAnalysisModifier::OOMetaClass::isApplicableTo(const DataCollection&
 }
 
 /******************************************************************************
-* Creates and initializes a computation engine that will compute the modifier's results.
-******************************************************************************/
-Future<AsynchronousModifier::EnginePtr> VoronoiAnalysisModifier::createEngine(const ModifierEvaluationRequest& request, const PipelineFlowState& input)
+ * Is called by the pipeline system before a new modifier evaluation begins.
+ ******************************************************************************/
+void VoronoiAnalysisModifier::preevaluateModifier(const ModifierEvaluationRequest& request, PipelineEvaluationResult::EvaluationTypes& evaluationTypes, TimeInterval& validityInterval) const
 {
-    // Get the input particles.
-    const Particles* particles = input.expectObject<Particles>();
+    // Indicate that we will do different computations depending on whether the pipeline is evaluated in interactive mode or not.
+    if(request.interactiveMode())
+        evaluationTypes = PipelineEvaluationResult::EvaluationType::Interactive;
+    else
+        evaluationTypes = PipelineEvaluationResult::EvaluationType::Noninteractive;
+}
+
+/******************************************************************************
+* Modifies the input data.
+******************************************************************************/
+Future<PipelineFlowState> VoronoiAnalysisModifier::evaluateModifier(const ModifierEvaluationRequest& request, PipelineFlowState&& state)
+{
+    // Get the current particle positions.
+    Particles* particles = state.expectMutableObject<Particles>();
     particles->verifyIntegrity();
+
+    // In interactive mode, fetch and return outdated results from the pipeline cache if available.
+    if(request.interactiveMode()) {
+        if(PipelineFlowState cachedState = request.modificationNode()->getCachedPipelineNodeOutput(request.time(), true)) {
+            if(const Particles* cachedParticles = cachedState.getObject<Particles>()) {
+                particles->tryToAdoptProperties(cachedParticles, {
+                    cachedParticles->getProperty(Particles::CoordinationProperty),
+                    cachedParticles->getProperty(QStringLiteral("Atomic Volume")),
+                    cachedParticles->getProperty(QStringLiteral("Cavity Radius")),
+                    computeIndices() ? cachedParticles->getProperty(QStringLiteral("Max Face Order")) : nullptr,
+                    computeIndices() ? cachedParticles->getProperty(QStringLiteral("Voronoi Index")) : nullptr,
+                }, {particles});
+            }
+            if(computePolyhedra()) {
+                if(const SurfaceMesh* cachedSurface = cachedState.getObjectBy<SurfaceMesh>(request.modificationNode(), QStringLiteral("voronoi-polyhedra"))) {
+                    state.addObject(cachedSurface);
+                }
+            }
+            // Adopt all global attributes computed by the modifier from the cached state.
+            state.adoptAttributesFrom(cachedState, request.modificationNode());
+        }
+        return std::move(state);
+    }
+
     const Property* posProperty = particles->expectProperty(Particles::PositionProperty);
 
     // Get simulation cell.
-    const SimulationCell* inputCell = input.expectObject<SimulationCell>();
+    const SimulationCell* inputCell = state.expectObject<SimulationCell>();
     if(inputCell->is2D())
         throw Exception(tr("The Voronoi modifier does not support 2d simulation cells."));
 
@@ -151,35 +184,42 @@ Future<AsynchronousModifier::EnginePtr> VoronoiAnalysisModifier::createEngine(co
         radii = particles->inputParticleRadii();
 
     // The Voro++ library uses 32-bit integers. It cannot handle more than 2^31 input points.
-    if(posProperty->size() > (size_t)std::numeric_limits<int>::max())
+    if(particles->elementCount() > (size_t)std::numeric_limits<int>::max())
         throw Exception(tr("Voronoi analysis modifier is limited to a maximum of %1 particles in the current program version.").arg(std::numeric_limits<int>::max()));
 
-    DataOORef<SurfaceMesh> polyhedraMesh;
+    SurfaceMesh* polyhedraMesh = nullptr;
     if(computePolyhedra()) {
         // Output the surface mesh representing the computed Voronoi polyhedra.
-        polyhedraMesh = DataOORef<SurfaceMesh>::create(ObjectInitializationFlag::DontCreateVisElement, tr("Voronoi polyhedra"));
-        polyhedraMesh->setIdentifier(input.generateUniqueIdentifier<SurfaceMesh>(QStringLiteral("voronoi-polyhedra")));
-        polyhedraMesh->setCreatedByNode(request.modificationNode());
+        polyhedraMesh = state.createObjectWithVis<SurfaceMesh>(QStringLiteral("voronoi-polyhedra"), request.modificationNode(), polyhedraVis(), tr("Voronoi polyhedra"));
         polyhedraMesh->setDomain(inputCell);
-        polyhedraMesh->setVisElement(polyhedraVis());
     }
 
     // Create engine object. Pass all relevant modifier parameters to the engine as well as the input data.
-    return std::make_shared<VoronoiAnalysisEngine>(
-            request,
-            input.stateValidity(),
-            particles,
+    auto engine = std::make_unique<VoronoiAnalysisEngine>(
+            request.modificationNode(),
+            particles->elementCount(),
             posProperty,
             selectionProperty,
             particles->getProperty(Particles::IdentifierProperty),
             std::move(radii),
             inputCell,
-            std::move(polyhedraMesh),
+            polyhedraMesh,
             computeIndices(),
             computeBonds(),
             edgeThreshold(),
             faceThreshold(),
-            relativeFaceThreshold());
+            relativeFaceThreshold(),
+            bondsVis());
+
+    // Perform the calculation in a separate thread.
+    return asyncLaunch([
+            state = std::move(state),
+            engine = std::move(engine)]() mutable
+    {
+        engine->perform();
+        engine->applyResults(state);
+        return std::move(state);
+    });
 }
 
 /******************************************************************************
@@ -187,13 +227,12 @@ Future<AsynchronousModifier::EnginePtr> VoronoiAnalysisModifier::createEngine(co
 ******************************************************************************/
 void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
 {
-    OVITO_ASSERT(_simCell);
-
-    setProgressText(tr("Performing Voronoi analysis"));
-    beginProgressSubSteps(_polyhedraMesh ? 2 : 1);
+    TaskProgress progress(this_task::ui());
+    progress.setText(tr("Performing Voronoi analysis"));
+    progress.beginSubSteps(_polyhedraMesh ? 2 : 1);
 
     // Compute total simulation cell volume.
-    _simulationBoxVolume = _simCell->volume3D();
+    _simulationBoxVolume = simCell().volume3D();
 
     // Stores the starting vertex index and the vertex count for each Voronoi polyhedron.
     std::vector<std::pair<SurfaceMesh::vertex_index, SurfaceMesh::size_type>> polyhedraVertices;
@@ -261,7 +300,7 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
         faceGrower.emplace(polyhedraMesh);
     }
 
-    // For generting the "Voronoi Order" bond property.
+    // For generating the "Voronoi Order" bond property.
     PropertyFactory<int32_t> bondVoronoiOrder;
     if(_computeBonds) {
         bondVoronoiOrder = PropertyFactory<int32_t>(Bonds::OOClass(), 0, QStringLiteral("Voronoi Order"));
@@ -282,7 +321,7 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
     }
 
     // The squared edge length threshold.
-    // Apply additional prefactor of 4, because Voronoi cell vertex coordinates are all scaled by factor of 2.
+    // Apply additional pre-factor of 4, because Voronoi cell vertex coordinates are all scaled by factor of 2.
     const FloatType sqEdgeThreshold = _edgeThreshold * _edgeThreshold * 4;
 
     // Prepare output data arrays.
@@ -354,9 +393,10 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
                     if(_polyhedraMesh) {
                         QMutexLocker locker(bondMutex);
                         meshFace = faceGrower->createFace(meshRegionIndex);
-                        BufferWriteAccess<int32_t, access_mode::write>{adjacentCellProperty, meshFace == 0}[meshFace] = neighbor_id;
+                        DataBuffer::BufferInitialization firstAccessMode = (meshFace == 0) ? DataBuffer::Uninitialized : DataBuffer::Initialized;
+                        BufferWriteAccess<int32_t, access_mode::write>{adjacentCellProperty, firstAccessMode}[meshFace] = neighbor_id;
                         if(_computeBonds)
-                            BufferWriteAccess<int64_t, access_mode::write>{faceBondIndexProperty, meshFace == 0}[meshFace] = -1;
+                            BufferWriteAccess<int64_t, access_mode::write>{faceBondIndexProperty, firstAccessMode}[meshFace] = -1;
                         polyhedraMesh.mutableTopology()->createEdge(meshVertexBaseIndex + i, meshVertexBaseIndex + k, meshFace);
                     }
 
@@ -397,8 +437,9 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
 
                     if(_polyhedraMesh) {
                         QMutexLocker locker(bondMutex);
-                        BufferWriteAccess<FloatType, access_mode::write>{faceAreaProperty, meshFace == 0}[meshFace] = area;
-                        BufferWriteAccess<int32_t, access_mode::write>{faceOrderProperty, meshFace == 0}[meshFace] = faceOrder;
+                        DataBuffer::BufferInitialization firstAccessMode = (meshFace == 0) ? DataBuffer::Uninitialized : DataBuffer::Initialized;
+                        BufferWriteAccess<FloatType, access_mode::write>{faceAreaProperty, firstAccessMode}[meshFace] = area;
+                        BufferWriteAccess<int32_t, access_mode::write>{faceOrderProperty, firstAccessMode}[meshFace] = faceOrder;
                     }
 
                     if((faceAreaThreshold == 0 || area > faceAreaThreshold) && faceOrder >= 3) {
@@ -416,13 +457,13 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
                             Vector3 diff = delta - normal;
                             Vector3I pbcShift = Vector3I::Zero();
                             for(size_t dim = 0; dim < 3; dim++) {
-                                if(_simCell->hasPbc(dim)) {
-                                    pbcShift[dim] = (int)std::round(_simCell->inverseMatrix().prodrow(diff, dim));
+                                if(simCell().hasPbc(dim)) {
+                                    pbcShift[dim] = (int)std::round(simCell().reciprocalCellMatrix().prodrow(diff, dim));
                                 }
                             }
 #ifdef OVITO_DEBUG
                             // Verify the computed pbc shift vector. The corrected neighbor vector should now align with the face normal vector.
-                            delta -= _simCell->reducedToAbsolute(pbcShift.toDataType<FloatType>());
+                            delta -= simCell().reducedToAbsolute(pbcShift.toDataType<FloatType>());
                             OVITO_ASSERT(std::abs(delta.dot(normal) / delta.length() / normal.length()) > 1.0 - FLOATTYPE_EPSILON);
 #endif
                             Bond bond = { index, (size_t)neighbor_id, pbcShift };
@@ -461,14 +502,14 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
     std::vector<size_t> voronoiBufferIndex;
 
     // Decide whether to use Voro++ container class or our own implementation.
-    if(_simCell->isAxisAligned()) {
+    if(simCell().isAxisAligned()) {
         // Use Voro++ container.
-        double ax = _simCell->matrix()(0,3);
-        double ay = _simCell->matrix()(1,3);
-        double az = _simCell->matrix()(2,3);
-        double bx = ax + _simCell->matrix()(0,0);
-        double by = ay + _simCell->matrix()(1,1);
-        double bz = az + _simCell->matrix()(2,2);
+        double ax = simCell().cellMatrix()(0,3);
+        double ay = simCell().cellMatrix()(1,3);
+        double az = simCell().cellMatrix()(2,3);
+        double bx = ax + simCell().cellMatrix()(0,0);
+        double by = ay + simCell().cellMatrix()(1,1);
+        double bz = az + simCell().cellMatrix()(2,2);
         if(ax > bx) std::swap(ax,bx);
         if(ay > by) std::swap(ay,by);
         if(az > bz) std::swap(az,bz);
@@ -482,7 +523,7 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
         if(!_radii) {
             // All particles have a uniform size.
             voro::container voroContainer(ax, bx, ay, by, az, bz, nx, ny, nz,
-                    _simCell->hasPbc(0), _simCell->hasPbc(1), _simCell->hasPbc(2), (int)std::ceil(voro::optimal_particles));
+                    simCell().hasPbc(0), simCell().hasPbc(1), simCell().hasPbc(2), (int)std::ceil(voro::optimal_particles));
 
             // Insert particles into Voro++ container.
             for(size_t index = 0; index < positionsArray.size(); index++) {
@@ -495,14 +536,13 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
             }
             if(!count) return;
 
-            setProgressMaximum(count);
+            progress.setMaximum(count);
 
             voro::c_loop_all cl(voroContainer);
             voro::voronoicell_neighbor v;
             if(cl.start()) {
                 do {
-                    if(!incrementProgressValue())
-                        return;
+                    progress.incrementValue();
                     if(!voroContainer.compute_cell(v,cl))
                         continue;
                     processCell(v, cl.pid(), voronoiBuffer, voronoiBufferIndex, nullptr);
@@ -519,7 +559,7 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
         else {
             // Particles have non-uniform sizes -> Compute polydisperse Voronoi tessellation.
             voro::container_poly voroContainer(ax, bx, ay, by, az, bz, nx, ny, nz,
-                    _simCell->hasPbc(0), _simCell->hasPbc(1), _simCell->hasPbc(2),
+                    simCell().hasPbc(0), simCell().hasPbc(1), simCell().hasPbc(2),
                     (int)std::ceil(voro::optimal_particles));
 
             // Insert particles into Voro++ container.
@@ -533,15 +573,15 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
                 count++;
             }
 
-            if(!count) return;
-            setProgressMaximum(count);
+            if(!count)
+                return;
+            progress.setMaximum(count);
 
             voro::c_loop_all cl(voroContainer);
             voro::voronoicell_neighbor v;
             if(cl.start()) {
                 do {
-                    if(!incrementProgressValue())
-                        return;
+                    progress.incrementValue();
                     if(!voroContainer.compute_cell(v,cl))
                         continue;
                     processCell(v, cl.pid(), voronoiBuffer, voronoiBufferIndex, nullptr);
@@ -559,42 +599,37 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
     else {
         // Special code path for non-orthogonal simulation cells:
 
-        // Prepare the nearest neighbor list generator.
-        NearestNeighborFinder nearestNeighborFinder;
-        if(!nearestNeighborFinder.prepare(positions(), _simCell, selection()))
-            return;
+        // Prepare the nearest neighbor finder.
+        NearestNeighborFinder nearestNeighborFinder(16, positions(), simCell(), selection());
 
         // This is the size we use to initialize Voronoi cells. Must be larger than the simulation box.
         double boxDiameter = sqrt(
-                  _simCell->matrix().column(0).squaredLength()
-                + _simCell->matrix().column(1).squaredLength()
-                + _simCell->matrix().column(2).squaredLength());
+                  simCell().cellMatrix().column(0).squaredLength()
+                + simCell().cellMatrix().column(1).squaredLength()
+                + simCell().cellMatrix().column(2).squaredLength());
 
         // The normal vectors of the three cell planes.
         std::array<Vector3,3> planeNormals;
-        planeNormals[0] = _simCell->cellNormalVector(0);
-        planeNormals[1] = _simCell->cellNormalVector(1);
-        planeNormals[2] = _simCell->cellNormalVector(2);
+        planeNormals[0] = simCell().cellNormalVector(0);
+        planeNormals[1] = simCell().cellNormalVector(1);
+        planeNormals[2] = simCell().cellNormalVector(2);
 
-        Point3 corner1 = Point3::Origin() + _simCell->matrix().column(3);
-        Point3 corner2 = corner1 + _simCell->matrix().column(0) + _simCell->matrix().column(1) + _simCell->matrix().column(2);
+        Point3 corner1 = simCell().cellOrigin();
+        Point3 corner2 = corner1 + simCell().cellMatrix().column(0) + simCell().cellMatrix().column(1) + simCell().cellMatrix().column(2);
 
         QMutex bondMutex;
         QMutex indexMutex;
         BufferReadAccess<GraphicsFloatType> radiusArray(_radii);
 
         // Perform analysis, particle-wise parallel.
-        setProgressMaximum(_positions->size());
-        parallelForChunksWithProgress(_positions->size(), [&](size_t startIndex, size_t chunkSize, ProgressingTask& operation) {
+        parallelForInnerOuter(_positions->size(), 4096, progress, [&](auto&& iterate) {
             std::vector<int> localVoronoiBuffer;
             std::vector<size_t> localVoronoiBufferIndex;
-            for(size_t index = startIndex; chunkSize--; index++) {
-                if(operation.isCanceled()) return;
-                if((index % 256) == 0) operation.incrementProgressValue(256);
+            iterate([&](size_t index) {
 
                 // Skip unselected particles (if requested).
                 if(selectionArray && selectionArray[index] == 0)
-                    continue;
+                    return;
 
                 // Build Voronoi cell.
                 voro::voronoicell_neighbor v;
@@ -605,7 +640,7 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
                 // Cut Voronoi cell at simulation cell boundaries in non-periodic directions.
                 bool skipParticle = false;
                 for(size_t dim = 0; dim < 3; dim++) {
-                    if(!_simCell->hasPbc(dim)) {
+                    if(!simCell().hasPbc(dim)) {
                         double r;
                         r = 2 * planeNormals[dim].dot(corner2 - positionsArray[index]);
                         if(r <= 0) skipParticle = true;
@@ -617,7 +652,7 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
                 }
                 // Skip particles that are located outside of non-periodic box boundaries.
                 if(skipParticle)
-                    continue;
+                    return;
 
                 // This function will be called for every neighbor particle.
                 int nvisits = 0;
@@ -639,15 +674,13 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
                 nearestNeighborFinder.visitNeighbors(nearestNeighborFinder.particlePos(index), visitFunc);
 
                 processCell(v, index, localVoronoiBuffer, localVoronoiBufferIndex, &bondMutex);
-            }
+            });
             if(!localVoronoiBufferIndex.empty()) {
                 QMutexLocker locker(&indexMutex);
                 voronoiBufferIndex.insert(voronoiBufferIndex.end(), localVoronoiBufferIndex.cbegin(), localVoronoiBufferIndex.cend());
                 voronoiBuffer.insert(voronoiBuffer.end(), localVoronoiBuffer.cbegin(), localVoronoiBuffer.cend());
             }
         });
-        if(isCanceled())
-            return;
     }
 
     if(maxFaceOrders()) {
@@ -664,6 +697,7 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
         OVITO_ASSERT(indexData == voronoiBuffer.cend());
         maxFaceOrdersArray.reset();
         voronoiIndicesArray.reset();
+        this_task::throwIfCanceled();
 
         // Re-use the output particle property as an output mesh region property.
         if(_polyhedraMesh) {
@@ -677,28 +711,28 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
 
     // Finalize the polyhedral mesh.
     if(_polyhedraMesh) {
-        nextProgressSubStep();
-        beginProgressSubStepsWithWeights({1,12,1,1,1});
+        progress.nextSubStep();
+        progress.beginSubSteps({1,12,1,1,1});
 
         // First, connect adjacent faces from the same Voronoi cell.
         polyhedraMesh.connectOppositeHalfedges();
 
         // The polyhedral cells should now be closed manifolds.
         OVITO_ASSERT(polyhedraMesh.topology()->isClosed());
-        nextProgressSubStep();
-        setProgressMaximum(polyhedraMesh.faceCount());
+        progress.nextSubStep();
+        progress.setMaximum(polyhedraMesh.faceCount());
 
         // Merge mesh vertices that are shared by adjacent Voronoi polyhedra.
 
         // Initialize disjoint set data structure to keep track which vertices have been merged with which.
         std::vector<SurfaceMesh::vertex_index> parents(polyhedraMesh.vertexCount());
         std::vector<SurfaceMesh::vertex_index> ranks(polyhedraMesh.vertexCount(), 0);
-        std::iota(parents.begin(), parents.end(), (SurfaceMesh::vertex_index)0);
+        boost::algorithm::iota(parents, (SurfaceMesh::vertex_index)0);
 
         // Iterate over all Voronoi faces.
         BufferReadAccess<int32_t> adjacentCellArray(adjacentCellProperty);
         for(SurfaceMesh::face_index face = 0; face < polyhedraMesh.faceCount(); face++) {
-            if(!setProgressValueIntermittent(face)) return;
+            progress.setValueIntermittent(face);
             SurfaceMesh::region_index region = faceGrower->faceRegion(face);
 
             // We know for each Voronoi face which Voronoi polyhedron is on the other side.
@@ -780,25 +814,25 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
             }
             while(edge != ffe);
         }
-        nextProgressSubStep();
+        progress.nextSubStep();
 
         // Transfer edges from vertices that are going to be deleted to remaining vertices.
         for(SurfaceMesh::edge_index edge = 0; edge < polyhedraMesh.edgeCount(); edge++) {
             SurfaceMesh::vertex_index new_vertex = parents[polyhedraMesh.vertex2(edge)];
             polyhedraMesh.transferFaceBoundaryToVertex(edge, new_vertex);
-            if(isCanceled()) return;
+            this_task::throwIfCanceled();
         }
-        nextProgressSubStep();
+        progress.nextSubStep();
 
         // Delete unused vertices.
         for(SurfaceMesh::vertex_index vertex = polyhedraMesh.vertexCount() - 1; vertex >= 0; vertex--) {
             if(parents[vertex] != vertex) {
                 vertexGrower->deleteVertex(vertex);
-                if(isCanceled()) return;
+                this_task::throwIfCanceled();
             }
         }
-        nextProgressSubStep();
-        setProgressMaximum(polyhedraMesh.faceCount());
+        progress.nextSubStep();
+        progress.setMaximum(polyhedraMesh.faceCount());
 
         BufferWriteAccess<int64_t, access_mode::read_write> faceBondIndices(faceBondIndexProperty);
 
@@ -806,8 +840,7 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
         for(SurfaceMesh::face_index face = 0; face < polyhedraMesh.faceCount(); face++) {
             if(polyhedraMesh.hasOppositeFace(face))
                 continue;
-            if(!setProgressValueIntermittent(face))
-                return;
+            progress.setValueIntermittent(face);
 
             // We know for each Voronoi face which Voronoi polyhedron is on the other side.
             SurfaceMesh::region_index adjacentRegion = adjacentCellArray[face];
@@ -820,7 +853,6 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
 
             SurfaceMesh::edge_index first_edge = polyhedraMesh.firstFaceEdge(face);
             SurfaceMesh::vertex_index vertex1 = polyhedraMesh.vertex1(first_edge);
-            SurfaceMesh::vertex_index vertex2 = polyhedraMesh.vertex2(first_edge);
 
             // Iterate over all edges/faces adjacent to one of the vertices.
             for(SurfaceMesh::edge_index edge = polyhedraMesh.firstVertexEdge(vertex1); edge != SurfaceMesh::InvalidIndex; edge = polyhedraMesh.nextVertexEdge(edge)) {
@@ -841,41 +873,36 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::perform()
             }
             OVITO_ASSERT(polyhedraMesh.hasOppositeFace(face) || polyhedraVertices[adjacentRegion].second == 0);
         }
+        vertexGrower.reset();
+        faceGrower.reset();
 
-        endProgressSubSteps();
+        progress.endSubSteps();
+
+#ifdef OVITO_DEBUG
+        polyhedraMesh.mesh()->verifyMeshIntegrity();
+#endif
     }
 
-    endProgressSubSteps();
-
-    // Release data that is no longer needed.
-    _positions.reset();
-    _selection.reset();
-    _particleIdentifiers.reset();
-    _simCell.reset();
-    _radii.reset();
+    progress.endSubSteps();
 }
 
 /******************************************************************************
 * Injects the computed results of the engine into the data pipeline.
 ******************************************************************************/
-void VoronoiAnalysisModifier::VoronoiAnalysisEngine::applyResults(const ModifierEvaluationRequest& request, PipelineFlowState& state)
+void VoronoiAnalysisModifier::VoronoiAnalysisEngine::applyResults(PipelineFlowState& state)
 {
-    VoronoiAnalysisModifier* modifier = static_object_cast<VoronoiAnalysisModifier>(request.modifier());
     Particles* particles = state.expectMutableObject<Particles>();
 
-    if(_inputFingerprint.hasChanged(particles))
-        throw Exception(tr("Cached modifier results are obsolete, because the number or the storage order of input particles has changed."));
-
     particles->createProperty(coordinationNumbers());
     particles->createProperty(atomicVolumes());
     particles->createProperty(cavityRadii());
 
-    if(modifier->computeIndices()) {
-        if(voronoiIndices())
-            particles->createProperty(voronoiIndices());
-        if(maxFaceOrders())
-            particles->createProperty(maxFaceOrders());
+    if(voronoiIndices())
+        particles->createProperty(voronoiIndices());
+    if(maxFaceOrders())
+        particles->createProperty(maxFaceOrders());
 
+    if(voronoiIndices() || maxFaceOrders()) {
         state.setStatus(PipelineStatus(PipelineStatus::Success,
             tr("Maximum face order: %1").arg(maxFaceOrder().load())));
     }
@@ -887,22 +914,18 @@ void VoronoiAnalysisModifier::VoronoiAnalysisEngine::applyResults(const Modifier
                         "This may be a result of particles being located outside of the simulation box boundaries. "
                         "See user manual for more information.\n"
                         "Simulation box volume: %1\n"
-                        "Voronoi cell volume sum: %2").arg(_simulationBoxVolume).arg(voronoiVolumeSum())));
+                        "Voronoi cell volume sum: %2").arg(_simulationBoxVolume).arg(voronoiVolumeSum().load())));
     }
 
-    if(modifier->computeBonds() && _computeBonds) {
+    if(_computeBonds) {
         // Insert output object into the pipeline.
         std::vector<PropertyPtr> bondProperties;
         if(_bondVoronoiOrder)
             bondProperties.push_back(_bondVoronoiOrder);
-        particles->addBonds(bonds(), modifier->bondsVis(), bondProperties);
+        particles->addBonds(bonds(), _bondsVis, bondProperties);
     }
 
-    // Output the surface mesh representing the computed Voronoi polyhedra.
-    if(_polyhedraMesh)
-        state.addObjectWithUniqueId<SurfaceMesh>(_polyhedraMesh);
-
-    state.addAttribute(QStringLiteral("Voronoi.max_face_order"), QVariant::fromValue(maxFaceOrder().load()), request.modificationNode());
+    state.addAttribute(QStringLiteral("Voronoi.max_face_order"), QVariant::fromValue(maxFaceOrder().load()), createdByNode());
 }
 
 }   // End of namespace
