@@ -1,6 +1,6 @@
 ////////////////////////////////////////////////////////////////////////////////////////
 //
-//  Copyright 2023 OVITO GmbH, Germany
+//  Copyright 2025 OVITO GmbH, Germany
 //
 //  This file is part of OVITO (Open Visualization Tool).
 //
@@ -27,7 +27,9 @@
 
 namespace Ovito {
 
-IMPLEMENT_OVITO_CLASS(Property);
+IMPLEMENT_CREATABLE_OVITO_CLASS(Property);
+OVITO_CLASSINFO(Property, "DisplayName", "Property");
+OVITO_CLASSINFO(Property, "ClassNameAlias", "PropertyObject");  // For backward compatibility with OVITO 3.9.2
 DEFINE_VECTOR_REFERENCE_FIELD(Property, elementTypes);
 DEFINE_PROPERTY_FIELD(Property, title);
 SET_PROPERTY_FIELD_LABEL(Property, elementTypes, "Element types");
@@ -37,19 +39,20 @@ SET_PROPERTY_FIELD_CHANGE_EVENT(Property, title, ReferenceEvent::TitleChanged);
 /******************************************************************************
 * Constructor creating an empty property array.
 ******************************************************************************/
-Property::Property(ObjectInitializationFlags flags) : DataBuffer(flags)
+void Property::initializeObject(ObjectInitializationFlags flags)
 {
+    DataBuffer::initializeObject(flags);
 }
 
 /******************************************************************************
 * Constructor allocating a property array with given size and data layout.
 ******************************************************************************/
-Property::Property(ObjectInitializationFlags flags, BufferInitialization init, size_t elementCount, int dataType, size_t componentCount, const QString& name, int type, QStringList componentNames) :
-    DataBuffer(flags, init, elementCount, dataType, componentCount, std::move(componentNames)),
-    _name(name),
-    _type(type)
+void Property::initializeObject(ObjectInitializationFlags flags, BufferInitialization init, size_t elementCount, int dataType, size_t componentCount, const QStringView name, int typeId, QStringList componentNames)
 {
-    setIdentifier(name);
+    DataBuffer::initializeObject(flags, init, elementCount, dataType, componentCount, std::move(componentNames));
+
+    _typeId = typeId;
+    setName(name.toString());
 }
 
 #ifdef OVITO_DEBUG
@@ -66,20 +69,16 @@ Property::~Property()
 ******************************************************************************/
 OORef<RefTarget> Property::clone(bool deepCopy, CloneHelper& cloneHelper) const
 {
-    OVITO_ASSERT(this->identifier() == this->name());
-
     // Let the base class create an instance of this class.
     OORef<Property> clone = static_object_cast<Property>(DataBuffer::clone(deepCopy, cloneHelper));
 
     // Copy internal data.
     prepareReadAccess();
-    clone->_type = _type;
-    clone->_name = _name;
-    OVITO_ASSERT(clone->identifier() == clone->name());
+    clone->_typeId = _typeId;
     finishReadAccess();
 
 #ifdef OVITO_USE_SYCL
-    if(isBeingAccessedFromPython()) {
+    if(isBeingAccessedExternally()) {
         // Force flush SYCL queue to complete the memcpy to the cloned data buffer now.
         // That's needed because Python code may be performing subsequent writes to the old memory buffer that will go unnoticed by
         // SYCL. We need to make sure these happen after the memcpy is completed, because a direct write access to the array should never
@@ -91,19 +90,6 @@ OORef<RefTarget> Property::clone(bool deepCopy, CloneHelper& cloneHelper) const
     return clone;
 }
 
-/******************************************************************************
-* Sets the property's name.
-******************************************************************************/
-void Property::setName(const QString& newName)
-{
-    if(newName == name())
-        return;
-
-    _name = newName;
-    setIdentifier(newName);
-    notifyTargetChanged(PROPERTY_FIELD(title));
-}
-
 /******************************************************************************
 * Returns the display title of this property object in the user interface.
 ******************************************************************************/
@@ -112,6 +98,19 @@ QString Property::objectTitle() const
     return title().isEmpty() ? name() : title();
 }
 
+/******************************************************************************
+* Is called when the value of a property of this object has changed.
+******************************************************************************/
+void Property::propertyChanged(const PropertyFieldDescriptor* field)
+{
+    DataBuffer::propertyChanged(field);
+
+    if(field == PROPERTY_FIELD(DataObject::identifier) && title().isEmpty()) {
+        // Since the identifier is the property's name, the property's title potentially changes too.
+        notifyDependents(ReferenceEvent::TitleChanged);
+    }
+}
+
 /******************************************************************************
 * Generates a human-readable string representation of the data object reference.
 ******************************************************************************/
@@ -139,8 +138,8 @@ void Property::saveToStream(ObjectSaveStream& stream, bool excludeRecomputableDa
     prepareReadAccess();
     try {
         stream.beginChunk(0x100);
-        stream << _name;
-        stream << _type;
+        stream << name();
+        stream << _typeId;
         stream.endChunk();
         finishReadAccess();
     }
@@ -155,13 +154,14 @@ void Property::saveToStream(ObjectSaveStream& stream, bool excludeRecomputableDa
 ******************************************************************************/
 void Property::loadFromStream(ObjectLoadStream& stream)
 {
+    QString name;
     if(stream.formatVersion() >= 30007) {
         DataBuffer::loadFromStream(stream);
 
         // Current file format:
         stream.expectChunk(0x100);
-        stream >> _name;
-        stream >> _type;
+        stream >> name;
+        stream >> _typeId;
         stream.closeChunk();
     }
     else {
@@ -171,13 +171,13 @@ void Property::loadFromStream(ObjectLoadStream& stream)
         // For backward compatibility with OVITO 3.3.5.
         stream.expectChunk(0x01);
         stream.expectChunk(0x02);
-        stream >> _name;
-        stream >> _type;
+        stream >> name;
+        stream >> _typeId;
         DataBuffer::loadFromStream(stream);
         stream.closeChunk();
     }
 
-    setIdentifier(name());
+    setIdentifier(name);
 }
 
 /******************************************************************************
@@ -190,8 +190,10 @@ bool Property::equals(const Property& other) const
     other.prepareReadAccess();
 
     bool result = [&]() {
-        if(this->type() != other.type()) return false;
-        if(this->type() == GenericUserProperty && this->name() != other.name()) return false;
+        if(this->typeId() != other.typeId())
+            return false;
+        if(this->typeId() == GenericUserProperty && this->name() != other.name())
+            return false;
         return true;
     }();
 
@@ -204,6 +206,27 @@ bool Property::equals(const Property& other) const
     return DataBuffer::equals(other);
 }
 
+/******************************************************************************
+* Creates an empty copy of this property object - without copying the stored
+* array data but cloning the metadata and list of element types.
+******************************************************************************/
+PropertyPtr Property::cloneWithoutData(size_t newSize, int overrideDataType) const
+{
+    UndoSuspender noUndo;
+
+    PropertyPtr clone = PropertyPtr::create(
+            ObjectInitializationFlag::DontInitializeObject, DataBuffer::Uninitialized, newSize, overrideDataType != 0 ? overrideDataType : this->dataType(),
+            this->componentCount(), this->name(), this->typeId(), this->componentNames());
+
+    clone->setVisElements(this->visElements());
+    clone->setElementTypes(this->elementTypes());
+    clone->setTitle(this->title());
+    clone->setCreatedByNode(this->createdByNode());
+    clone->setEditableProxy(this->editableProxy());
+
+    return clone;
+}
+
 /******************************************************************************
 * Helper method that remaps the existing type IDs to a contiguous range starting at the given
 * base ID. This method is mainly used for file output, because some file formats
@@ -317,17 +340,19 @@ void Property::sortElementTypesById()
 /******************************************************************************
 * Creates an editable proxy object for this DataObject and synchronizes its parameters.
 ******************************************************************************/
-void Property::updateEditableProxies(PipelineFlowState& state, ConstDataObjectPath& dataPath) const
+void Property::updateEditableProxies(PipelineFlowState& state, ConstDataObjectPath& dataPath, bool forceProxyReplacement) const
 {
-    DataBuffer::updateEditableProxies(state, dataPath);
+    DataBuffer::updateEditableProxies(state, dataPath, forceProxyReplacement);
 
     // Note: 'this' may no longer exist at this point, because the base method implementation may
     // have already replaced it with a mutable copy.
     const Property* self = static_object_cast<Property>(dataPath.back());
 
-    if(Property* proxy = static_object_cast<Property>(self->editableProxy())) {
+    if(self->editableProxy() && !forceProxyReplacement) {
+        Property* proxy = static_object_cast<Property>(self->editableProxy());
+
         // Synchronize the actual data object with the editable proxy object.
-        OVITO_ASSERT(proxy->type() == self->type());
+        OVITO_ASSERT(proxy->typeId() == self->typeId());
         OVITO_ASSERT(proxy->dataType() == self->dataType());
         OVITO_ASSERT(proxy->title() == self->title());
 
@@ -342,7 +367,7 @@ void Property::updateEditableProxies(PipelineFlowState& state, ConstDataObjectPa
     else if(!self->elementTypes().empty()) {
         // Create and initialize a new proxy property object.
         // Note: We avoid copying the property data here by constructing the proxy Property from scratch instead of cloning the original data object.
-        OORef<Property> newProxy = OORef<Property>::create(ObjectInitializationFlag::DontCreateVisElement, DataBuffer::Uninitialized, 0, self->dataType(), self->componentCount(), self->name(), self->type(), self->componentNames());
+        OORef<Property> newProxy = OORef<Property>::create(ObjectInitializationFlag::DontCreateVisElement, DataBuffer::Uninitialized, 0, self->dataType(), self->componentCount(), self->name(), self->typeId(), self->componentNames());
         newProxy->setTitle(self->title());
 
         // Adopt the proxy objects corresponding to the element types, which have already been created by
@@ -370,7 +395,7 @@ const ElementType* Property::addNumericType(const PropertyContainerClass& contai
     // If the caller did not specify an element type class, let the PropertyConatiner class
     // determine the right element type class for the given property.
     if(elementTypeClass == nullptr) {
-        elementTypeClass = containerClass.typedPropertyElementClass(type());
+        elementTypeClass = containerClass.typedPropertyElementClass(typeId());
         if(elementTypeClass == nullptr)
             elementTypeClass = &ElementType::OOClass();
     }
@@ -381,7 +406,7 @@ const ElementType* Property::addNumericType(const PropertyContainerClass& contai
     // Second initialization phase for element types, which takes into account the assigned ID and name and the property type.
     elementType->setNumericId(id);
     elementType->setName(name);
-    elementType->initializeType(PropertyReference(&containerClass, this));
+    elementType->initializeType(OwnerPropertyRef(&containerClass, this));
 
     // Log in type name assigned by the caller as default value for the element type.
     // This is needed for the Python code generator to detect manual changes subsequently made by the user.
@@ -410,10 +435,10 @@ QString Property::nameWithComponent(int vectorComponent) const
 }
 
 /******************************************************************************
-* Throws an exception with an informative text if the given name is not a
+* Throws an exception with an informative text if the given string is not a
 * valid name for an OVITO property.
 ******************************************************************************/
-void Property::throwIfInvalidPropertyName(const QString& name)
+void Property::throwIfInvalidPropertyName(const QStringView name)
 {
     if(name.isEmpty())
         throw Exception(tr("Invalid empty property name. OVITO property names must have at least length 1."));
@@ -427,10 +452,32 @@ void Property::throwIfInvalidPropertyName(const QString& name)
         throw Exception(tr("Invalid property name: '%1'. OVITO property names must not start with whitespace.").arg(name));
     if(name.endsWith(QChar(' ')))
         throw Exception(tr("Invalid property name: '%1'. OVITO property names must not end with whitespace.").arg(name));
+    if(name.endsWith(QChar('_')))
+        throw Exception(tr("Invalid property name: '%1'. OVITO property names must not end with an underscore.").arg(name));
 }
 
 /******************************************************************************
-* Performs name mangeling if necessary to turn the given name into a valid property name.
+* Throws an exception with an informative text if the given string is not a
+* valid name for an OVITO property vector component.
+******************************************************************************/
+void Property::throwIfInvalidPropertyComponentName(const QStringView name)
+{
+    if(name.isEmpty())
+        throw Exception(tr("Invalid empty component name. OVITO vector component names must have at least length 1."));
+    if(name.contains(QChar('.')))
+        throw Exception(tr("Invalid component name: '%1'. Dots are not allowed in OVITO vector component names.").arg(name));
+    if(name.contains(QChar('/')))
+        throw Exception(tr("Invalid component name: '%1'. '/' is not allowed in OVITO vector component names.").arg(name));
+    if(name.contains(QChar(':')))
+        throw Exception(tr("Invalid component name: '%1'. ':' is not allowed in OVITO vector component names.").arg(name));
+    if(name.contains(QChar(' ')))
+        throw Exception(tr("Invalid component name: '%1'. OVITO vector component names must not contain whitespace.").arg(name));
+    if(name.endsWith(QChar('_')))
+        throw Exception(tr("Invalid component name: '%1'. OVITO vector component names must not end with an underscore.").arg(name));
+}
+
+/******************************************************************************
+* Performs name mangling if necessary to turn the given name into a valid property name.
 ******************************************************************************/
 QString Property::makePropertyNameValid(const QString& name)
 {
@@ -438,6 +485,27 @@ QString Property::makePropertyNameValid(const QString& name)
     mangledName.replace(QChar('.'), QChar('_'));
     mangledName.replace(QChar('/'), QChar('_'));
     mangledName.replace(QChar(':'), QChar('_'));
+    // Remove all underscores from the end of the name, because they interfere with
+    // the "underscore notation" in OVITO's Python API (PropertyContainer dict lookup).
+    while(mangledName.endsWith(QChar('_')))
+        mangledName.chop(1);
+    return mangledName;
+}
+
+/******************************************************************************
+* Performs name mangling if necessary to turn the given name into a valid vector property component name.
+******************************************************************************/
+QString Property::makeComponentNameValid(const QString& name)
+{
+    QString mangledName = name.trimmed();
+    mangledName.replace(QChar('.'), QChar('_'));
+    mangledName.replace(QChar('/'), QChar('_'));
+    mangledName.replace(QChar(':'), QChar('_'));
+    mangledName.replace(QChar(' '), QChar('_'));
+    // Remove all underscores from the end of the name, because they interfere with
+    // the "underscore notation" in OVITO's Python API (PropertyContainer dict lookup).
+    while(mangledName.endsWith(QChar('_')))
+        mangledName.chop(1);
     return mangledName;
 }
 
